$script:items = @()

Function Pack($config) {
    $id = $config["id"]
    Check-Id $id "pack"
    $depends = Parse-Depends $config["depends"]
    Append-Item @{category="pack"; id=$id; depends=$depends}
}

Function File($config) {
    $id = $config["id"]
    Check-Id $id "file"
    $content = $config["content"]
    $depends = Parse-Depends $config["depends"]
    Append-Item @{category="file"; id=$id; content=$content; depends=$depends}
}

Function Append-Item($item) {
    if (Contains $script:items $item) { throw "duplicate definition of $($item.category) id: ${($item.id)}"}
    $script:items = $script:items + $item
}

Function Check-Id($id, $category) {
    if (-not $id) { throw "Missing id for one $category." }
}

Function Parse-Depends($dependsStringOrArray) {
    if (-not $dependsStringOrArray) { return @() }

    if ($dependsStringOrArray.GetType().Name -eq "String") {
        $matched = "$dependsStringOrArray" -match "(pack|file|service|command)\(([\w\.\-\:\\\ \(\)]+)\)"
        if (-not $matched) { throw }
        return @{category=$matches[1]; id=$matches[2]}
    }
    $dependsStringOrArray | % { Parse-Depends $_ }  
}

Function Generate-ExecutionPlan($items) {
    $ep = New-Object System.Collections.Generic.List``1[System.Object]
    $iteratePath = New-Object System.Collections.Generic.List``1[System.Object]
    Generate-ExecutionPlanRecur $items $items $ep $iteratePath
    return $ep
}

Function Generate-ExecutionPlanRecur($items, $allItems, $ep, $iteratePath) {
   $items | % {
       $item = $_
       Check-CycleDepend $iteratePath $item
       $iteratePath.Add($item)
       if ($item.depends) {
           $item.depends | ? { Depend-IsDefined $allItems $_ } | % { Generate-ExecutionPlanRecur  (MapTo-Item $allItems $_) $allItems  $ep $iteratePath}
       }
       $iteratePath.RemoveAt(($iteratePath.Count - 1))  
       if (-not (Contains $ep $item)) { $ep.Add($item) }
   }
}

Function MapTo-Item($items, $index) {
    $items | ? { ($_.id -eq $index.id) -and ($_.category -eq $index.category)}
}

Function Check-CycleDepend($iteratePath, $item) {
    $cycleDepends = Contains $iteratePath $item
    if ($cycleDepends) { throw } 
}

Function Depend-IsDefined($items, $item) {
    $contained = Contains $items $item
    if ($contained) { return $true }
    throw "{ $($item.category) $({$item.id}) } not defined."
}

Function Contains($items, $item) {
    $founded =  $items | ? { $_.id -eq $item.id } | ? { $_.category -eq $item.category }
    return ($founded -ne $nil)
}

Function Apply-Spec($executionPlan) {
    Write-Host "Applying spec"
    $executionPlan | % {
        if ($_.category -eq "pack") {
            Write-Host Installing $_.id
            iex "cinst $($_.id)"
        }        
        if ($_.category -eq "file") {
            Write-Host "creating file $($_.id)"
            if (-not (Test-Path $_.id)) { New-Item -type file $_.id }
            Set-Content $_.id $_.content
        }
    } 
}
